#include "nvfs.h"

void nvfs_free_inode(struct nvfs_superblock *nvs, unsigned long i)
{
	unsigned long group, offset;
	struct nvfs_group *g;
	struct nvfs_group_info *gi;
	struct nvfs_inode *nvi;

	nvi = p_inode(nvs, i);
	nvfs_free_run(nvs, &nvi->i_xattr_block, 1, false);

	group = i >> nvs->log2_inodes_per_group;
	offset = i & ((1UL << nvs->log2_inodes_per_group) - 1);

	if (unlikely(group >= nvs->n_groups)) {
		nvfs_error(nvs, "freeing invalid inode %lu", i);
		return;
	}

	g = p_group(nvs, group);
	gi = &nvs->group_info[group];

	spin_lock(&gi->lock);

	if (!__test_and_clear_bit_le(offset, g->bitmap)) {
		nvfs_error(nvs, "inode bit %lu already cleared", i);
		spin_unlock(&gi->lock);
		return;
	}

	arch_wb_cache_pmem(g->bitmap + offset / 8, 1);

	gi->n_free_inodes++;
	nv_store_cached(g->n_free_inodes, cpu_to_le32(gi->n_free_inodes));

	spin_unlock(&gi->lock);

	percpu_counter_inc(&nvs->free_inodes_cnt);
}

static unsigned long inode_try_group(struct nvfs_superblock *nvs, unsigned long group, unsigned long offset, unsigned long limit)
{
	struct nvfs_group *g;
	struct nvfs_group_info *gi;
	unsigned long o;

	if (unlikely(WARN_ON(group >= nvs->n_groups))) {
		group = 0;
	}

	g = p_group(nvs, group);
	gi = &nvs->group_info[group];

	if (unlikely(READ_ONCE(gi->n_free_inodes)) < limit)
		return ~0;

search_again:
	o = nvfs_find_bit_in_bmp(g->bitmap, offset, 1UL << nvs->log2_inodes_per_group);
	if (likely(o != ~0)) {
		spin_lock(&gi->lock);

		if (unlikely(__test_and_set_bit_le(o, g->bitmap))) {
			/* we have lost the race */
			spin_unlock(&gi->lock);
			goto search_again;
		}
		arch_wb_cache_pmem(g->bitmap + o / 8, 1);

		gi->n_free_inodes--;
		nv_store_cached(g->n_free_inodes, cpu_to_le32(gi->n_free_inodes));
		pmem_wmb();
		spin_unlock(&gi->lock);
		percpu_counter_dec(&nvs->free_inodes_cnt);
	}

	return o;
}

struct inode *nvfs_new_inode(struct inode *dir, umode_t mode, dev_t rdev, const struct qstr *qstr)
{
	struct nvfs_superblock *nvs = dir->i_sb->s_fs_info;
	unsigned long group, offset, o, i, ino;
	struct nvfs_per_cpu *npc = raw_cpu_ptr(nvs->npc);
	struct nvfs_group *g;
	struct inode *inode;
	struct nvfs_inode *nvi;
	struct timespec64 now;
	unsigned i_flag;
	int r;

	ino = READ_ONCE(npc->current_inode);
	group = ino >> nvs->log2_inodes_per_group;
	offset = ino & ((1UL << nvs->log2_inodes_per_group) - 1);

	cond_resched();
	o = inode_try_group(nvs, group, offset, 1);
	if (unlikely(o == ~0)) {
		unsigned long x;
		uint64_t freei = percpu_counter_read_positive(&nvs->free_inodes_cnt);
		uint64_t average;

		average = div64_u64(freei, nvs->n_groups);

		offset = nvfs_get_random(1UL << nvs->log2_inodes_per_group);
		for (x = 0; x < nvs->n_groups; x++) {
			cond_resched();
			group = nvfs_get_random(nvs->n_groups);
			o = inode_try_group(nvs, group, offset, average);
			if (o != ~0)
				goto found;
		}
		for (group = 0; group < nvs->n_groups; group++) {
			cond_resched();
			o = inode_try_group(nvs, group, offset, 1);
			if (o != ~0)
				goto found;
		}
		r = -ENOSPC;
		goto ret0;
	}

found:
	g = p_group(nvs, group);

	ino = (group << nvs->log2_inodes_per_group) | o;

	if (unlikely(ino + 1 < nvs->n_groups << nvs->log2_inodes_per_group))
		WRITE_ONCE(npc->current_inode, ino + 1);
	else
		WRITE_ONCE(npc->current_inode, 0);

	nvi = p_inode(nvs, ino);

	nv_store_cached(nvi->i_size, cpu_to_le64(0));
	nv_store_cached(nvi->i_blocks, cpu_to_le64(0));
	for (i = 0; i < NVFS_DIRECT_BLOCKS; i++)
		nv_store_cached(nvi->i_direct_blocks[i], cpu_to_le64(0));
	for (i = 0; i < NVFS_INDIRECT_BLOCKS; i++)
		nv_store_cached(nvi->i_indirect_blocks[i], cpu_to_le64(0));

	nv_store_cached(nvi->i_xattr_block, cpu_to_le64(0));
	nv_store_cached(nvi->i_xattr[0].x, nvfs_xattr_entry_make_size(1U << nvs->log2_inode_size) - offsetof(struct nvfs_inode, i_xattr));

	inode = new_inode(dir->i_sb);
	if (!inode) {
		r = -ENOMEM;
		goto ret1;
	}

	inode->i_ino = ino;

	inode_init_owner(inode, dir, mode);

	if (unlikely(S_ISBLK(inode->i_mode) || unlikely(S_ISCHR(inode->i_mode)))) {
		nv_store_cached(nvi->i_rdev, cpu_to_le64(huge_encode_dev(rdev)));
	}

	nv_store_cached(nvi->i_links, cpu_to_le32(1));
	i_flag = i_to_nmi(dir)->i_flag & NVFS_FLAGS_INHERITED;
	nv_store_cached(nvi->i_flag, cpu_to_le32(i_flag));

	now = current_time(inode);
	nv_store_cached(nvi->i_ctime, cpu_to_le64(now.tv_sec));
	nv_store_cached(nvi->i_mtime, cpu_to_le64(now.tv_sec));
	nv_store_cached(nvi->i_atime, cpu_to_le64(now.tv_sec));
	nv_store_cached(nvi->i_btime, cpu_to_le64(now.tv_sec));
	nv_store_cached(nvi->i_ctime_nsec, cpu_to_le32(now.tv_nsec));
	nv_store_cached(nvi->i_mtime_nsec, cpu_to_le32(now.tv_nsec));
	nv_store_cached(nvi->i_atime_nsec, cpu_to_le32(now.tv_nsec));
	nv_store_cached(nvi->i_btime_nsec, cpu_to_le32(now.tv_nsec));

	nvfs_fill_inode(inode, nvi);

	if (insert_inode_locked(inode) < 0) {
		nvfs_error(nvs, "inode %lu already used", ino);
		r = -EUCLEAN;
		goto ret2;
	}

	r = dquot_initialize(inode);
	if (unlikely(r))
		goto ret3;

	r = dquot_alloc_inode(inode);
	if (unlikely(r))
		goto ret3;

	r = nvfs_init_acl(inode, dir);
	if (unlikely(r))
		goto ret4;

	r = nvfs_init_security(inode, dir, qstr);
	if (unlikely(r))
		goto ret4;

	nv_store_cached(nvi->i_uid, cpu_to_le32(i_uid_read(inode)));
	nv_store_cached(nvi->i_gid, cpu_to_le32(i_gid_read(inode)));
	nv_store_cached(nvi->i_mode, cpu_to_le32(inode->i_mode));

	nv_flush_cached(nvi, offsetof(struct nvfs_inode, i_xattr[0].data));

	return inode;

ret4:
	dquot_free_inode(inode);
ret3:
	inode->i_flags |= S_NOQUOTA;
	clear_nlink(inode);
	discard_new_inode(inode);
	return ERR_PTR(r);
ret2:
	make_bad_inode(inode);
	iput(inode);
ret1:
	nvfs_free_inode(nvs, i);
ret0:
	return ERR_PTR(r);
}
