#include "nvfs.h"

static const struct inode_operations nvfs_link_inode_operations = {
	.get_link = simple_get_link,
	.getattr = nvfs_getattr,
	.setattr = nvfs_setattr,
	.update_time = nvfs_update_time,
	.listxattr = nvfs_listxattr,
	.get_acl = nvfs_get_acl,
	.set_acl = nvfs_set_acl,
};

static const struct inode_operations nvfs_special_inode_operations = {
	.getattr = nvfs_getattr,
	.setattr = nvfs_setattr,
	.update_time = nvfs_update_time,
	.listxattr = nvfs_listxattr,
	.get_acl = nvfs_get_acl,
	.set_acl = nvfs_set_acl,
};

const struct address_space_operations nvfs_aops = {
	.direct_IO = nvfs_rw_iter_locked,
	.readpage = nvfs_cache_readpage,
	.writepage = nvfs_cache_writepage,
	.write_begin = nvfs_cache_write_begin,
	.write_end = nvfs_cache_write_end,
};

void nvfs_set_inode_flags(struct inode *inode)
{
	unsigned flags, iflags;

	flags = i_to_nmi(inode)->i_flag;
	iflags = inode->i_flags & ~(S_SYNC | S_APPEND | S_IMMUTABLE | S_NOATIME | S_DIRSYNC);

	if (unlikely(flags & FS_SYNC_FL))
		iflags |= S_SYNC;
	if (unlikely(flags & FS_APPEND_FL))
		iflags |= S_APPEND;
	if (unlikely(flags & FS_IMMUTABLE_FL))
		iflags |= S_IMMUTABLE;
	if (unlikely(flags & FS_NOATIME_FL))
		iflags |= S_NOATIME;
	if (unlikely(flags & FS_DIRSYNC_FL))
		iflags |= S_DIRSYNC;

	inode->i_flags = iflags;
}

void nvfs_fill_inode(struct inode *inode, struct nvfs_inode *nvi)
{
	struct nvfs_superblock *nvs = inode->i_sb->s_fs_info;
	struct nvfs_memory_inode *nmi = i_to_nmi(inode);

	nmi->nvi = nvi;
	nmi->was_mmapped = false;

	inode->i_size = le64_to_cpu(nvi->i_size);
	inode->i_blocks = le64_to_cpu(nvi->i_blocks);
	inode->i_ctime.tv_sec = le64_to_cpu(nvi->i_ctime);
	inode->i_mtime.tv_sec = le64_to_cpu(nvi->i_mtime);
	inode->i_atime.tv_sec = le64_to_cpu(nvi->i_atime);
	nmi->i_btime.tv_sec = le64_to_cpu(nvi->i_btime);
	inode->i_ctime.tv_nsec = le32_to_cpu(nvi->i_ctime_nsec);
	inode->i_mtime.tv_nsec = le32_to_cpu(nvi->i_mtime_nsec);
	inode->i_atime.tv_nsec = le32_to_cpu(nvi->i_atime_nsec);
	nmi->i_btime.tv_nsec = le32_to_cpu(nvi->i_btime_nsec);
	set_nlink(inode, le32_to_cpu(nvi->i_links));
	nmi->i_flag = le32_to_cpu(nvi->i_flag);
	nvfs_set_inode_flags(inode);

	if (likely(S_ISREG(inode->i_mode))) {
		inode->i_op = &nvfs_file_inode_operations;
		inode->i_fop = &nvfs_file_fops;
		if (nvs->mount_flags & NVFS_DAX)
			inode->i_flags |= S_DAX;
		inode->i_mapping->a_ops = &nvfs_aops;
	} else if (S_ISDIR(inode->i_mode)) {
		inode->i_op = &nvfs_dir_inode_operations;
		inode->i_fop = &nvfs_dir_fops;
	} else if (S_ISLNK(inode->i_mode)) {
		inode->i_op = &nvfs_link_inode_operations;
		if (!(nmi->i_flag & NVFS_FLAG_FAST_SYMLINK))
			inode->i_link = nvs->memory_map + le64_to_cpu(nvi->i_direct_blocks[0]);
		else
			inode->i_link = nvi->i_fast_symlink;
	} else {
		inode->i_op = &nvfs_special_inode_operations;
		init_special_inode(inode, inode->i_mode, huge_decode_dev(le64_to_cpu(nvi->i_rdev)));
	}
}

struct inode *nvfs_iget(struct super_block *sb, unsigned long ino)
{
	struct inode *inode;
	struct nvfs_inode *nvi;
	struct nvfs_memory_inode *nmi;
	struct nvfs_superblock *nvs;

	inode = iget_locked(sb, ino);
	if (unlikely(!inode))
		return ERR_PTR(-ENOMEM);
	if (!(inode->i_state & I_NEW))
		return inode;

	nvs = sb->s_fs_info;
	nvi = p_inode(nvs, ino);

	nmi = i_to_nmi(inode);

	i_uid_write(inode, le32_to_cpu(nvi->i_uid));
	i_gid_write(inode, le32_to_cpu(nvi->i_gid));
	inode->i_mode = le32_to_cpu(nvi->i_mode);

	nvfs_fill_inode(inode, nvi);

	unlock_new_inode(inode);

	return inode;
}

void nvfs_evict_inode(struct inode *inode)
{
	bool want_delete = false;
	if (!inode->i_nlink && !is_bad_inode(inode)) {
		want_delete = true;
		dquot_initialize(inode);
	} else {
		dquot_drop(inode);
	}

	truncate_inode_pages_final(&inode->i_data);

	if (want_delete) {
		if (nvfs_inode_has_indirect_blocks(inode))
			nvfs_punch_pages(i_to_nmi(inode), 0, -1, false);
		if (S_ISDIR(inode->i_mode))
			nvfs_delete_directory(inode);
	}

	clear_inode(inode);

	if (want_delete) {
		dquot_free_inode(inode);
	        dquot_drop(inode);
		nvfs_free_inode(inode->i_sb->s_fs_info, inode->i_ino);
	}
}

int nvfs_getattr(const struct path *path, struct kstat *stat, uint32_t request_mask, unsigned int query_flags)
{
	struct inode *inode = d_inode(path->dentry);
	struct nvfs_memory_inode *nmi = i_to_nmi(inode);

	if (request_mask & STATX_BTIME) {
		stat->result_mask |= STATX_BTIME;
		stat->btime.tv_sec = nmi->i_btime.tv_sec;
		stat->btime.tv_nsec = nmi->i_btime.tv_nsec;
	}

	if (nmi->i_flag & FS_COMPR_FL)
		stat->attributes |= STATX_ATTR_COMPRESSED;
	if (nmi->i_flag & FS_IMMUTABLE_FL)
		stat->attributes |= STATX_ATTR_IMMUTABLE;
	if (nmi->i_flag & FS_APPEND_FL)
		stat->attributes |= STATX_ATTR_APPEND;
	if (nmi->i_flag & FS_NODUMP_FL)
		stat->attributes |= STATX_ATTR_NODUMP;
	if (inode->i_flags & S_DAX)
		stat->attributes |= STATX_ATTR_DAX;
	stat->attributes_mask =
		STATX_ATTR_COMPRESSED |
		STATX_ATTR_IMMUTABLE |
		STATX_ATTR_APPEND |
		STATX_ATTR_NODUMP |
		STATX_ATTR_DAX;

	generic_fillattr(inode, stat);

	return 0;
}

int nvfs_setattr(struct dentry *dentry, struct iattr *attr)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(d_inode(dentry));
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	struct nvfs_inode *nvi = nmi->nvi;
	int r;

	r = setattr_prepare(dentry, attr);
	if (unlikely(r))
		return r;

	if (is_quota_modification(&nmi->vfs_inode, attr)) {
		r = dquot_initialize(&nmi->vfs_inode);
		if (unlikely(r))
			return r;
	}

	setattr_copy(&nmi->vfs_inode, attr);

	if (attr->ia_valid & ATTR_UID) {
		nv_store(nvi->i_uid, cpu_to_le32(i_uid_read(&nmi->vfs_inode)));
	}
	if (attr->ia_valid & ATTR_GID) {
		nv_store(nvi->i_gid, cpu_to_le32(i_gid_read(&nmi->vfs_inode)));
	}
	if (attr->ia_valid & ATTR_CTIME) {
		nv_store(nvi->i_ctime, cpu_to_le64(attr->ia_ctime.tv_sec));
		nv_store(nvi->i_ctime_nsec, cpu_to_le32(attr->ia_ctime.tv_nsec));
	}
	if (attr->ia_valid & ATTR_MTIME) {
		nv_store(nvi->i_mtime, cpu_to_le64(attr->ia_mtime.tv_sec));
		nv_store(nvi->i_mtime_nsec, cpu_to_le32(attr->ia_mtime.tv_nsec));
	}
	if (attr->ia_valid & ATTR_ATIME) {
		nv_store(nvi->i_atime, cpu_to_le64(attr->ia_atime.tv_sec));
		nv_store(nvi->i_atime_nsec, cpu_to_le32(attr->ia_atime.tv_nsec));
	}
	if (attr->ia_valid & ATTR_MODE) {
		nv_store(nvi->i_mode, cpu_to_le32(nmi->vfs_inode.i_mode));
	}

	if (attr->ia_valid & ATTR_SIZE) {
		WARN_ON(!inode_is_locked(&nmi->vfs_inode));
		if (attr->ia_size != nmi->vfs_inode.i_size) {
			r = nvfs_punch_hole(nmi, attr->ia_size, LLONG_MAX - ((1UL << nvs->log2_page_size) - 1));
			if (unlikely(r))
				return r;
			nvfs_set_inode_size(nmi, attr->ia_size);
		}
	}

	return 0;
}

int nvfs_update_time(struct inode *inode, struct timespec64 *time, int flags)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(inode);
	struct nvfs_inode *nvi = nmi->nvi;
	struct timespec64 now;

	if (!time) {
		now = current_time(inode);
		time = &now;
	}

	if (flags & S_CTIME) {
		nmi->vfs_inode.i_ctime = *time;
		nv_store(nvi->i_ctime, cpu_to_le64(time->tv_sec));
		nv_store(nvi->i_ctime_nsec, cpu_to_le32(time->tv_nsec));
	}
	if (flags & S_MTIME) {
		nmi->vfs_inode.i_mtime = *time;
		nv_store(nvi->i_mtime, cpu_to_le64(time->tv_sec));
		nv_store(nvi->i_mtime_nsec, cpu_to_le32(time->tv_nsec));
	}
	if (flags & S_ATIME) {
		nmi->vfs_inode.i_atime = *time;
		nv_store(nvi->i_atime, cpu_to_le64(time->tv_sec));
		nv_store(nvi->i_atime_nsec, cpu_to_le32(time->tv_nsec));
	}

	return 0;
}
